% This file simulates Judd's (1985) two-class model
% Ludwig Straub and Ivan Werning, 2019

% This file requires interp1qr.m and qinterp2.m


%% housekeeping
clear all
close all
if 1==0
        %matlabpool close
        dcluster=parcluster;
        dcluster.ResourceTemplate='-l nodes=^N^,software=MATLAB_Distrib_Comp_Engine+^N^,walltime=24:00:00,mem=128gb';
        dcluster.saveProfile;
        parpool open 4
end


%% Choose which figure to produce data for:
figno = 1;  % 1, 2, 3. (4 = test)

if figno == 1
    % loop over sigma around 1
    PARLIST = [0.75 0.9 0.95 0.99 1.025 1.05 1.1 1.25];
elseif figno == 2
    % loop over sigma just below 1
    PARLIST = [0.5 0.525 0.55 0.575 0.6 0.625 0.65 0.675 0.7 0.725 0.75 0.775 0.8...
        0.81 0.82 0.825 0.83 0.84 0.85 0.86 0.87 0.875 0.88 0.89 ...
        0.9 0.905 0.91 0.915 0.92 0.925 0.93 0.935 0.94 0.945 0.95 0.955 0.96 0.961 0.962 0.963 0.964 0.965 0.967 0.968 0.969 0.97];
elseif figno == 3
    % loop over gammahat = gamma relative to zero-tax marginal utilities
    PARLIST = [-0.8:0.2:0.8];
elseif figno == 4
    PARLIST = 1.25; % Just run the default parameters with the value here as sigma
end


%%% Use smaller grid?
small_grid = 0;


%% Big Loop

for param=PARLIST

    disp([' **** Fig ' num2str(figno) ', param = ' num2str(param) ' **** '])
    
    clearvars -EXCEPT SIGMA KLOW GAMMAHAT REDIST ITER_SELF_G param figno small_grid
    close all
    
    
    %% Set parameters
    
    % Grid size (need large grid for accuracy, especially if redist > 0
    % where upper bound of k is larger)
    ngridk      = 512;  % 512; % 1024;
    ngridC      = 300;  % 300; % 600;
    if figno == 2, ngridC = 200; ngridk = 256; end  % use slightly less precise grid for speed of convergence plot
    
    if small_grid==1, ngridC = 100; ngridk = 128; end 
    
    % No of iterations for self-generating set (1 is sufficient)
    iter_self_g = 10;
    
    % No of iterations of Bellman (default = 200)
    iter_Bell   = 200;
    
    % Crop factor
    crop_factor = 0.001;
    
    % Set inverse EIS
    sigma       = 1.25;
    if figno == 1 || figno == 2 || figno == 4, sigma = param; end
    sigma_w     = sigma;
    
    % Other parameters
    beta        = 0.95;
    alpha       = 0.3 ;  % share of capital
    delta       = .1;  % depreciation
    g_over_y    = 0.20;  % set g as % of zero tax steady state output

    
    %% Compute auxiliary variables
    kgolden     = (delta/alpha)^(1/(alpha-1));
    kstar1      = ((1/alpha)*(1/beta-1+delta))^(1/(alpha-1)); % zero tax steady state
    kstar2      = ((1/alpha)*(1/beta^2-1+delta))^(1/(alpha-1)); % log steady state?
    kmax        = delta^(1/(alpha-1));
    
    y_zero_tax  = kstar1^alpha;
    g           = g_over_y * y_zero_tax;
    k_sub_g     = fzero(@(k) k^alpha + (1-delta)*k - k/beta - g,[0 kstar1]);
    k_up_g      = fzero(@(k) k^alpha + (1-delta)*k- k/beta -g ,[kstar1 kmax]);
    C_up_g      = (1-beta)/beta * k_up_g;
    C_sub_g     = (1-beta)/beta * k_sub_g;
    
    % compute zero tax steady state objects
    Cstar1      = (1-beta)/beta * kstar1;
    cstar1      = kstar1^alpha + (1-delta) * kstar1 - Cstar1 - kstar1 - g;
    
    % Set gamma (i.e. preferences for redistribution)
    gammahat = 0; redist = 1;
    Up_over_up  = (((1-alpha)/alpha*((1-beta)/beta + delta))*beta/(1-beta))^sigma; % upsilon at zero tax steady state
    gammastar   = (Up_over_up + 1)^-1;
    gamma       = gammahat*gammastar;
    if figno == 3
        redist = param;
        Cstar_redist= Cstar1*(1 - redist);
        cstar_redist= cstar1 + redist*Cstar1;   
        gamma       = ( (Cstar_redist)^(-sigma) / (cstar_redist)^(-sigma_w) + 1)^(-1);
    end

    % utility functions
    u           = @(c) c.^(1-sigma_w)/(1-sigma_w);
    U           = @(c) c.^(1-sigma)/(1-sigma);
     
    
    %% set grids
    if sigma<1
        klow    = kstar2/2;  % converge to zero tax s.s. anyway
        khigh   = 3;
    else
        klow    =  k_sub_g*(1+1e-5); %klow_bound*(1+1e-5);  % allow for different bound with tax bound.
        khigh   =  3;
    end
    
    % For redist < 0, use different, much greater upper bound
    if figno == 3 && redist < 0
        khigh       = 10;  % Alternative: k_up_g*(1-1e-5)
    end
    
    % For figure 2, use tighter bounds on k to improve accuracy
    if figno == 2
        klow = 1.3129;
    end
    
    kk          = linspace(klow,khigh,ngridk)';
    KK          = (kk*ones(1,ngridC))';
    
    psi_C       = linspace(0,1,ngridC)';
    PSI_C       = psi_C*ones(1,ngridk);
    
    Clow        = (1-beta)/beta *kk;
    Chigh       = Clow;
    CC          = cell(iter_self_g+2,1);
    CC{1}       = (1-psi_C)*Clow' + psi_C*Chigh';
    
    ngridc     = 20000;
    psi_c      = linspace(0,1,ngridc);
    
    
    %% 1. Self generation iterations
    disp('1. Computing the self-generating set...')
    CCC=CC{1}*0; XX =CC{1}*0;
    Cmin=Clow*0;Cmax=Chigh*0;
    
    cstar   = klow^alpha + (1-delta)*klow - klow/beta - g;
    clow    = 1e-9;
    clbound = min(clow,cstar); % only add states with c>=clbound in the s-g construction below

    
    for j=1:iter_self_g
        for i=1:ngridk
            CCC        = CC{j}.*(beta*(CC{j}+KK)/kk(i)).^(-1/sigma);
            XX         = (kk(i).^alpha+(1-delta)*kk(i)) -CC{j} -KK - g;
            
            if (XX>=clbound) == zeros(size(XX>=clbound))
                disp('control c>=c_lbound infeasible numerically. Check k bounds, especially klow');
            end
            
            Cmin(i,1)  = min(CCC(XX>=clbound));
            Cmax(i,1)  = max(CCC(XX>=clbound));
        end
        CC{j+1}        = (1-psi_C)*Cmin' + psi_C*Cmax' ;
    end

    % tiny crop to pass the test
    Cmin_crop = Cmin*(1-crop_factor) + Cmax*crop_factor;
    Cmax_crop = Cmax*(1-crop_factor) + Cmin*crop_factor;
    CC{j+2}    = (1-psi_C)*Cmin_crop' + psi_C*Cmax_crop';
    CC = CC{j+2};
     
    
    %% 2. Check self generating set and construct intervals for control cmin and cmax
    disp('2. Checking self-generating set...')
    c_theory_flag = CC*0; % matrix which shows the points for which we used theory to determine a succeeding state
    w_up_g = @(k) (k<k_up_g) .* (1-beta)/beta .*k + C_up_g * (k>=k_up_g) .* (k/k_up_g).^(1/sigma);

    cmax=CC*0;cmin=CC*0;
    for i= 1:ngridk
        cmax_inner  = zeros(ngridC,1); cmin_inner  = zeros(ngridC,1);
        for j= 1:ngridC

            ccc     = psi_c*(kk(i).^alpha+(1-delta)*kk(i)-g);
            CCC     = CC(j,i)*(kk(i)/beta/(kk(i).^alpha+(1-delta)*kk(i)-g)./(1-psi_c)).^(1/(-sigma));
            kkk     = (kk(i).^alpha+(1-delta)*kk(i))-CCC-ccc-g;

            cindices         = find((CCC>=interp1(kk,CC(1,:),kkk)) & (interp1(kk,CC(end,:),kkk)>=CCC) & (kkk<=khigh) & (kkk>=klow));
            
            if isempty(cindices) == 1
                disp('c index has empty set. Guess c from theory instead...')

                c_theory_flag(j,i) = 1;
                % make case distinction to compute a theoretically
                % possible control
                if CC(j,i)<=w_up_g(kk(i))
                     % can jump to a steady state (k_ss, C_ss) in 1 step
                     k_ss     = ( beta * CC(j,i)/(1-beta))^(sigma/(sigma-1)) * kk(i)^(-1/(sigma-1));
                     C_ss     = (1-beta)/beta * k_ss;
                     c_ss     = kk(i)^alpha + (1-delta)*kk(i) - (C_ss/CC(j,i))^sigma * kk(i)/beta -g;
                     c_guess = c_ss;
                     C_guess = C_ss;
                     k_guess = k_ss;
                % now we should use c = 0 but that might not work if
                % khigh < k_up_g, so use the lowest c that gets you to
                % k' = khigh
                elseif kkk(1)>khigh
                    % need to solve k' = khigh for c
                    fun      = @(x) x - CC(j,i) * (beta*x/kk(i)).^(1/sigma) - khigh;
                    xbar     = CC(j,i)^(sigma/(sigma-1)) * ( beta / kk(i) )^(1/(sigma-1));
                    x_guess  = fzero(fun,[xbar kk(i)^alpha + (1-delta)*kk(i) ]);
                    k_guess  = khigh;
                    C_guess  = x_guess - k_guess;
                    c_guess  = kk(i)^alpha + (1-delta)*kk(i) - x_guess -g;

                    % check whether there are numerical problems and if
                    % so define new guess
                    check2 = interp1(kk,CC(end,:),k_guess)>=C_guess;

                    if check2 == 0
                        C_guess = interp1(kk,CC(end,:),khigh);
                        x_guess = khigh + C_guess;
                        k_guess = khigh;
                        c_guess  = kk(i)^alpha + (1-delta)*kk(i)- x_guess -g;
                    end
                end

                % check whether c_guess is indeed inside the s-g set
                check1 = C_guess>=interp1(kk,CC(1,:),k_guess);
                check2 = interp1(kk,CC(end,:),k_guess)>=C_guess;
                check3 = k_guess<=khigh;
                check4 = k_guess>=klow;

                if check1 + check2 + check3 + check4 < 4
                    disp('The guessed state is not inside the s-g set.');
                end


                cmax_inner(j,1)  = c_guess;  cmin_inner(j,1)  = c_guess;
            else

                %%% CASE cindices IS NONEMPTY
                % CHECK FOR GAPS IN cindices (might happen if ngridk
                % low)

                if max(cindices) - min(cindices) > length(cindices) -1
                    %disp('the set of feasible c has a numerical gap.')
                    % find the values for k for which Cmax > CCC > Cmin
                    CC_interp = interp1(kkk,CCC,kk');
                    % select the kk indices for which (kkk,CCC)
                    % actually intersects the s-g set
                    kselection = (CC(1,:)<= CC_interp) & (CC(end,:)>=CC_interp) ...
                        + [(CC_interp(1:end-1)>=CC(1,1:end-1)) & (CC_interp(2:end)<CC(1,2:end)) 0];

                    % compute the indices for which (kkk,CCC) in the
                    % interior
                    kk_indices = find(kselection>0);


                    if max(kk_indices) - min(kk_indices) > length(kk_indices) - 1
                        disp('weird')
                    end

                    max_kk_indices = max(kk_indices)+1; % start looking for suitable (kkk,CCC) 1 node to the right of when they are inside.
                    min_kk_indices = min(kk_indices);

                    selection      = ( (kk(min_kk_indices) <= kkk(cindices)) & (kk(max_kk_indices) >= kkk(cindices)) );

                    cindices_new   = cindices(selection);

                    cindices       = cindices_new;

                    if max(cindices) - min(cindices) > length(cindices) -1
                        disp('there is still a gap.')
                    end

                end


                cmax_inner(j,1)  = ccc(max(cindices)); cmin_inner(j,1)= ccc(min(cindices));
            end

        end
        cmax(:,i)  = cmax_inner ; cmin(:,i)  = cmin_inner;
        
    end
    
    %% 3. Iterate the Bellman equation
    disp('3. Iterating the Bellman equation...')
    
    if iter_self_g == 1
        disp('... using Vguess as go to ray in one step ... ')
        % Guess value function: go to ray in one step
        KKguess	= ((1-beta)^(sigma) * beta^(-sigma)*CC.^(-sigma).*KK).^(1/(1-sigma));
        CCguess = (1-beta)/beta*KKguess;
        ccguess = KK.^alpha +(1-delta)*KK - CCguess - KKguess-g;
        ccprimeguess = KKguess.^alpha +(1-delta)*KKguess-g - CCguess - KKguess;

        % make c's close to 0 but negative  equal to 0.000001
        cc_zero                    = (ccguess > -1e-10) + (ccguess <=0);
        ccguess(cc_zero == 2)      = 0.000001;
        cc_zero                    = (ccprimeguess>-1e-10) + (ccprimeguess <= 0);
        ccprimeguess(cc_zero == 2) = 0.000001;

        Vguess  = (1-gamma)*u(ccguess)+gamma*U(CCguess) + beta * (gamma*U(CCguess) + (1-gamma)*u(ccprimeguess))/(1-beta);
    else
        disp('... using Vguess as steady state value given k (equal across C) ... ')
        % guess steady state value corresponding to k
        ccguess = KK.^alpha +(1-delta)*KK - KK/beta-g;
        CCguess = (1-beta)/beta * KK;
        Vguess  = ((1-gamma)*u(ccguess)+gamma*U(CCguess))/(1-beta);
    end
    
    V = Vguess; %zeros(size(Vguess)); %Vguess
    
    if isreal(Vguess)==0
        disp('Houston, we have a problem')
    end
    
    Cfun    = @(c,k,Cminus) Cminus*(k/beta/(k.^alpha+(1-delta)*k-g-c)).^(1/(-sigma));
    kfun    = @(c,k,Cminus) (k.^alpha+(1-delta)*k)-Cfun(c,k,Cminus)-c-g;
    psifun  = @(c,k,Cminus) (Cfun(c,k,Cminus)-interp1qr(kk,CC(1,:)',kfun(c,k,Cminus)))./(interp1qr(kk,CC(end,:)',kfun(c,k,Cminus)) - interp1qr(kk,CC(1,:)', kfun(c,k,Cminus)));    
    
    c_opt = ccguess;
    Vnew=V*0;
    
    c_opt_new = V*0;
    clear Vguess ccguess CCguess KKguess ccprimeguess
    disp('iteration __ max(abs(V-Vnew)) __ avg(abs(V-Vnew)) __ max(abs(policies))')
    
    inf_flag = V*0;
    
    for iter=1:iter_Bell
        tic
        for i=1:ngridk
            for j=1:ngridC
                
                chi = 1e-12; % helps getting qinterp2 points interior
                [c_opt_new(j,i), bla] = fminbnd(@(c) -(1-gamma)*u(c) -gamma*U(Cfun(c,KK(j,i),CC(j,i))) - beta * qinterp2(KK , PSI_C , V, kfun(c,KK(j,i),CC(j,i))*(1-chi) + chi*(khigh/2 + klow/2), psifun(c,KK(j,i),CC(j,i))*(1-chi) + chi*0.5), cmin(j,i), cmax(j,i));
                Vnew(j,i)=-bla;
                
                if isnan(bla)==1
                    disp('hola. NaN appeared in the value function.');
                end
                if isinf(bla)==1
                    disp('salut. Inf appeared in the value function.');
                    inf_flag(j,i) = 1;
                end
            end
        end
        crit=max(max(abs(Vnew-V)));
        crit_policy=max(max(abs(c_opt_new-c_opt)));
        avg_crit = mean(mean(abs(Vnew-V)));
        disp([iter crit avg_crit crit_policy])
        disp(['There were ' num2str(sum(sum(inf_flag))) ' infinities in V'])
        V=Vnew;
        c_opt=c_opt_new;
        toc
    end
    
    
    %% Compute and plot policy functions
    disp('4. Computing plots...')
    policy_c = c_opt;
    policy_C= zeros(ngridC,ngridk); policy_k= policy_C; policy_psi= policy_C;
    parfor i=1:ngridk
        for j=1:ngridC
            policy_C(j,i)   = Cfun(policy_c(j,i),KK(j,i),CC(j,i));
            policy_k(j,i)   = kfun(policy_c(j,i),KK(j,i),CC(j,i));
            policy_psi(j,i) = psifun(policy_c(j,i),KK(j,i),CC(j,i));
        end
    end
    
    figure(2); quiver(KK,CC, policy_k-KK,policy_C-CC,0.5); grid on
    title('Phase diagram in (k, C) space')
    xlabel('k'); ylabel('C');
    figure(3); surf(KK,CC,V); title('Value function over k and C'); xlabel('k'); ylabel('C');
    
    
    %% phase diagram
    figure(4); 
    quiver(KK,psi_C*ones(1,ngridk),policy_k-KK,policy_psi - psi_C*ones(1,ngridk),4);
    title('Phase diagram in (k, psi) space')
    xlabel('k'); ylabel('psi');
    grid on; hold on
    
    psi_steady = (kk*(1-beta)/beta-CC(1,:)')./(CC(end,:)'-CC(1,:)');
    [~, ind] = max(V); 
    psi_max = psi_C(ind);

    plot(kk, psi_steady,'b')
    plot(kk, psi_max   ,'k')

    
    %% add optimal path
    T=1000; ktime=zeros(T,1); psitime=ktime; VC_time = zeros(T-1,1); Vk_time = zeros(T-1,1);
    
    ktime(1,1)   = kstar1;
    psitime(1,1) = fminbnd(@(psiC) -interp2(kk,psi_C, V, ktime(1,1), psiC),0,1);

    for t=1:T-1
        psitime(t+1) = interp2(kk,psi_C, policy_psi, ktime(t),psitime(t));
        ktime(t+1)   = interp2(kk,psi_C, policy_k, ktime(t),psitime(t));
    end
    figure(4);plot(ktime,psitime,'r+-');
    Ctime = interp2(KK, psi_C*ones(1,ngridk),CC,ktime,psitime);
    
    legend('Phase diagram', 'Steady state curve', 'Optimal V for given k', 'Optimal path')
        

    %% time series
    T=1000;
    figure(5); plot(ktime); grid on
    title('Evolution of capital k')
    xlabel('time (years)')
    xlim([0 300])
    
    figure(6); plot(1-(Ctime(1:T-1)+ktime(2:T))./(alpha*ktime(1:T-1).^alpha+(1-delta)*ktime(1:T-1)),'b');
    hold on; grid on
    plot(alpha*ktime(1:T-1).^(alpha-1)+(1-delta),'r');
    plot((Ctime(1:T-1)+ktime(2:T))./ktime(1:T-1)  ,'k');
    title('Evolution of taxes and interest rates')
    legend('taxes T', 'Post-tax interest rate R', 'Pre-tax interest rate Rstar')
    xlabel('time (years)')
    xlim([0 300])
    
    
    %% save workspace
    disp('4. Saving the workspace...')
    
    path = ['fig' num2str(figno) '/'];
    filename = num2str(['judd_','sigma_',num2str(sigma),'redist',num2str(redist),...
        'sigma_w_',num2str(sigma_w),'_grid',num2str(ngridk),'x',num2str(ngridC),...
        'range_',num2str(klow),'-',num2str(khigh),...
        'g_', num2str(g),...
        'iter_sg_',num2str(iter_self_g),...
        '.mat']);
    disp(filename); save([path filename])
    
    if figno <= 3
        close all
    end
    
    
end % back to param loop